@@ -1,6 +1,6 @@
 /*@z02.c:Lexical Analyser:Declarations@***************************************/
 /*                                                                           */
-/*  THE LOUT DOCUMENT FORMATTING SYSTEM (VERSION 3.08)                       */
+/*  THE LOUT DOCUMENT FORMATTING SYSTEM (VERSION 3.10)                       */
 /*  COPYRIGHT (C) 1991, 1996 Jeffrey H. Kingston                             */
 /*                                                                           */
 /*  Jeffrey H. Kingston (jeff@cs.usyd.edu.au)                                */
@@ -51,6 +51,7 @@ static	FULL_CHAR	*frst;		/* address of first buffer character */
 static	FULL_CHAR	*limit;		/* just past last char in buffer     */
 static	FULL_CHAR	*buf;		/* the character buffer start pos    */
 static	int		blksize;	/* size of block read; others too    */
+static	FULL_CHAR	last_char;	/* last char read in from file       */
 static	FULL_CHAR	*startline;	/* position in buff of last newline  */
 static	FILE_NUM	this_file;	/* number of currently open file     */
 static	FILE		*fp;		/* current input file                */
@@ -58,6 +59,8 @@ static	FILE_POS	file_pos;	/* current file position             */
 static	short		ftype;		/* the type of the current file      */
 static	OBJECT		next_token;	/* next token if already read	     */
 static	int		offset;		/* where to start reading in file    */
+static	int		first_line_num;	/* number of first line (if offset)  */
+static	BOOLEAN		same_file;	/* TRUE if same file as preceding    */
 static	FULL_CHAR	*mem_block;	/* file buffer                       */
 
 static int stack_free;		/* first free slot in lexical stack  */
@@ -67,6 +70,7 @@ static struct {
   FULL_CHAR	*limit;		/* just past last char in buffer     */
   FULL_CHAR	*buf;		/* the character buffer start pos    */
   int		blksize;	/* size of block read; others too    */
+  FULL_CHAR	last_char;	/* last char read in from file	     */
   FULL_CHAR	*startline;	/* position in buff of last newline  */
   FILE_NUM	this_file;	/* number of currently open file     */
   FILE		*fp;		/* current input file                */
@@ -74,6 +78,9 @@ static struct {
   short		ftype;		/* the type of the current file      */
   OBJECT	next_token;	/* next token if already read	     */
   int		offset;		/* where to start reading in file    */
+  int		first_line_num;	/* number of first line (if offset)  */
+  BOOLEAN	same_file;	/* TRUE if same file as preceding    */
+  long		save_ftell;	/* ftell() position if same_file     */
   FULL_CHAR	*mem_block;	/* file buffer                       */
 } lex_stack[MAX_LEX_STACK];
 
@@ -165,7 +172,7 @@ void LexInit(void)
 
 /*@::LexPush(), LexPop()@*****************************************************/
 /*                                                                           */
-/*  LexPush(x, offs, ftype)                                                  */
+/*  LexPush(x, offs, ftype, lnum, same)                                      */
 /*                                                                           */
 /*  Start reading from the file sequence whose first file is x (subsequent   */
 /*  files are obtained from NextFile).  The first file (x) is to be fseeked  */
@@ -178,33 +185,42 @@ void LexInit(void)
 /*      INCLUDE_FILE   include file, must pop lexical analyser and continue  */
 /*      FILTER_FILE    filter file, return @End @FilterOut                   */
 /*                                                                           */
+/*  lnum is the line number at offs, to be used when creating file pos's     */
+/*  in the tokens returned.  same is TRUE when this file is the same as      */
+/*  the file currently being read, in which case there is no need to         */
+/*  close that file and open this one; just an fseek is required.            */
+/*                                                                           */
 /*****************************************************************************/
 
-void LexPush(FILE_NUM x, int offs, int ftyp)
+void LexPush(FILE_NUM x, int offs, int ftyp, int lnum, BOOLEAN same)
 {
-  debug3(DLA, D, "LexPush(%s, %d, %s)", FileName(x), offs,
-    ftyp==SOURCE_FILE ? "source" : ftyp==INCLUDE_FILE ? "include":"database");
+  debug5(DLA, D, "LexPush(%s, %d, %s, %d, %s)", FileName(x), offs,
+    ftyp==SOURCE_FILE ? "source" : ftyp==INCLUDE_FILE ? "include":"database",
+    lnum, bool(same));
   if( stack_free >= MAX_LEX_STACK - 1 )
   { if( ftyp == INCLUDE_FILE )
       Error(2, 1, "include file %s too deeply nested",
-        FATAL, PosOfFile(x), FileName(x));
+        FATAL, PosOfFile(x), FullFileName(x));
     else
       Error(2, 2, "database file %s too deeply nested",
         FATAL, PosOfFile(x), FileName(x));
   }
   if( stack_free >= 0 )  /* save current state */
-  { lex_stack[stack_free].chpt		= chpt;
-    lex_stack[stack_free].frst		= frst;
-    lex_stack[stack_free].limit		= limit;
-    lex_stack[stack_free].buf		= buf;
-    lex_stack[stack_free].blksize	= blksize;
-    lex_stack[stack_free].startline	= startline;
-    lex_stack[stack_free].this_file	= this_file;
-    lex_stack[stack_free].fp		= fp;
-    lex_stack[stack_free].ftype		= ftype;
-    lex_stack[stack_free].next_token	= next_token;
-    lex_stack[stack_free].offset	= offset;
-    lex_stack[stack_free].mem_block	= mem_block;
+  { lex_stack[stack_free].chpt		 = chpt;
+    lex_stack[stack_free].frst		 = frst;
+    lex_stack[stack_free].limit		 = limit;
+    lex_stack[stack_free].buf		 = buf;
+    lex_stack[stack_free].blksize	 = blksize;
+    lex_stack[stack_free].last_char	 = last_char;
+    lex_stack[stack_free].startline	 = startline;
+    lex_stack[stack_free].this_file	 = this_file;
+    lex_stack[stack_free].fp		 = fp;
+    lex_stack[stack_free].ftype		 = ftype;
+    lex_stack[stack_free].next_token	 = next_token;
+    lex_stack[stack_free].offset	 = offset;
+    lex_stack[stack_free].first_line_num = first_line_num;
+    lex_stack[stack_free].same_file	 = same_file;
+    lex_stack[stack_free].mem_block	 = mem_block;
     FposCopy( lex_stack[stack_free].file_pos, file_pos );
   }
   stack_free += 1;
@@ -213,11 +229,19 @@ void LexPush(FILE_NUM x, int offs, int ftyp)
   mem_block = (FULL_CHAR *) malloc((MAX_LINE+BUFFER_SIZE+2)*sizeof(FULL_CHAR));
   if( mem_block == NULL )
     Error(2, 3, "run out of memory when opening file %s",
-      FATAL, PosOfFile(x), FileName(x));
+      FATAL, PosOfFile(x), FullFileName(x));
   buf = chpt = &mem_block[MAX_LINE];
+  last_char = CH_NEWLINE;
   this_file = x;  offset = offs;
+  first_line_num = lnum;  same_file = same;
   ftype = ftyp;  next_token = nilobj;
-  *chpt = '\0';  fp = null;
+  *chpt = '\0';
+  if( same_file )
+  { lex_stack[stack_free-1].save_ftell = ftell(fp);
+  }
+  else
+  { fp = null;
+  }
 } /* end LexPush */
 
 
@@ -230,24 +254,32 @@ void LexPush(FILE_NUM x, int offs, int ftyp)
 void LexPop(void)
 { debug0(DLA, D, "LexPop()");
   assert( stack_free > 0, "LexPop: stack_free <= 0!" );
-  if( fp != null )  fclose(fp);
   stack_free--;
+  if( same_file )
+  { fseek(fp, lex_stack[stack_free].save_ftell, SEEK_SET);
+  }
+  else
+  { if( fp != null )  fclose(fp);
+  }
   ifdebug(DMA, D,
    DebugRegisterUsage(MEM_LEX,-1,-(MAX_LINE+BUFFER_SIZE+2)* (int) sizeof(FULL_CHAR))
   );
   free( (char *) mem_block);
-  mem_block    = lex_stack[stack_free].mem_block;
-  chpt         = lex_stack[stack_free].chpt;
-  frst         = lex_stack[stack_free].frst;
-  limit        = lex_stack[stack_free].limit;
-  buf          = lex_stack[stack_free].buf;
-  blksize      = lex_stack[stack_free].blksize;
-  startline    = lex_stack[stack_free].startline;
-  this_file    = lex_stack[stack_free].this_file;
-  fp           = lex_stack[stack_free].fp;
-  ftype        = lex_stack[stack_free].ftype;
-  next_token   = lex_stack[stack_free].next_token;
-  offset       = lex_stack[stack_free].offset;
+  mem_block	 = lex_stack[stack_free].mem_block;
+  chpt		 = lex_stack[stack_free].chpt;
+  frst		 = lex_stack[stack_free].frst;
+  limit		 = lex_stack[stack_free].limit;
+  buf		 = lex_stack[stack_free].buf;
+  blksize	 = lex_stack[stack_free].blksize;
+  last_char	 = lex_stack[stack_free].last_char;
+  startline	 = lex_stack[stack_free].startline;
+  this_file	 = lex_stack[stack_free].this_file;
+  fp		 = lex_stack[stack_free].fp;
+  ftype		 = lex_stack[stack_free].ftype;
+  next_token	 = lex_stack[stack_free].next_token;
+  offset	 = lex_stack[stack_free].offset;
+  first_line_num = lex_stack[stack_free].first_line_num;
+  same_file	 = lex_stack[stack_free].same_file;
   FposCopy( file_pos, lex_stack[stack_free].file_pos );
 } /* end LexPop */
 
@@ -319,6 +351,14 @@ static void srcnext(void)
     }
     chpt = frst;
     blksize = fread( (char *) buf, sizeof(char), BUFFER_SIZE, fp);
+    if( blksize > 0 )
+      last_char = *(buf + blksize - 1);
+    if( blksize < BUFFER_SIZE && last_char != CH_NEWLINE )
+    {
+      /* at end of file since blksize = 0; so add missing newline char */
+      blksize++;
+      last_char = *(buf+blksize-1) = CH_NEWLINE;
+    }
     debugcond4(DLA, DD, stack_free <= 1,
       "srcnext: %d = fread(0x%x, %d, %d, fp)",
       blksize, buf, sizeof(char), BUFFER_SIZE);
@@ -366,9 +406,16 @@ OBJECT LexGetToken(void)
   {
       case ESCAPE:
       
-	col_num(file_pos) = (startpos = p-1) - startline;
-	Error(2, 6, "character %c outside quoted string",
-	  WARN, &file_pos, *startpos);
+	if( ftype==DATABASE_FILE && *p>='a' && *p<='z' && *(p+1) == '{' /*}*/ )
+	{ res = NewToken(LBR, &file_pos, 0, 0, (unsigned) *p, StartSym);
+	  p += 2;
+	}
+	else
+	{
+	  col_num(file_pos) = (startpos = p-1) - startline;
+	  Error(2, 6, "character %c outside quoted string",
+	    WARN, &file_pos, *startpos);
+	}
 	break;
 
 
@@ -376,7 +423,19 @@ OBJECT LexGetToken(void)
       
 	debug1(DLA, DDD, "LexGetToken%s: comment", EchoFilePos(&file_pos));
 	while( (c = *p++) != CH_NEWLINE && c != '\0' );
-	--p;
+	if( c == CH_NEWLINE )
+	{
+	  /* do NEWLINE action, only preserve existing horizontal space */
+	  /* and don't count the newline in the vcount. */
+	  chpt = p;  srcnext();
+	  line_num(file_pos)++;
+	  col_num(file_pos) = 0;
+	  startline = (p = chpt) - 1;
+	}
+	else
+	{
+	  --p;
+	}
 	break;
 
 
@@ -405,32 +464,38 @@ OBJECT LexGetToken(void)
 
       case ENDFILE:
       
-	/* close current file, if any */
 	debug0(DLA, DDD, "LexGetToken: endfile");
-	if( fp != null )
-	{ fclose(fp);  fp = null;
-	  this_file = ftype == SOURCE_FILE ? NextFile(this_file) : NO_FILE;
+	if( !same_file )
+	{
+	  /* close current file, if any */
+	  if( fp != null )
+	  { fclose(fp);  fp = null;
+	    this_file = ftype == SOURCE_FILE ? NextFile(this_file) : NO_FILE;
+	  }
+
+	  /* open next file */
+	  while( this_file != NO_FILE )
+	  { file_num(file_pos) = this_file;
+	    line_num(file_pos) = 1;
+	    col_num(file_pos) = 0;
+	    fp = OpenFile(this_file, FALSE, TRUE);
+	    if( fp != null )  break;
+	    Error(2, 7, "cannot open file %s",
+	      WARN, &file_pos, FullFileName(this_file));
+	    this_file = ftype == SOURCE_FILE ? NextFile(this_file) : NO_FILE;
+	  }
 	}
 
-	/* open next file */
-	while( this_file != NO_FILE )
-	{ file_num(file_pos) = this_file;
-	  line_num(file_pos) = 1;
-	  col_num(file_pos) = 0;
-	  fp = OpenFile(this_file, FALSE, TRUE);
-	  if( fp != null )  break;
-	  Error(2, 7, "cannot open file %s",
-	    WARN, &file_pos, FileName(this_file));
-	  this_file = ftype == SOURCE_FILE ? NextFile(this_file) : NO_FILE;
-	}
 	if( fp != null )
 	{ if( offset != 0 )
-	  { debugcond1(DLA, DD, stack_free <= 1, "fseek(fp, %d, 0)", offset);
-	    fseek(fp, (long) offset, 0);
+	  { debugcond1(DLA, DD, stack_free <= 1, "fseek(fp, %d, SEEK_SET)", offset);
+	    fseek(fp, (long) offset, SEEK_SET);
 	    offset = 0L;
+	    line_num(file_pos) = first_line_num;
 	  }
 	  frst = limit = chpt = buf;
-	  blksize = 0;  srcnext();
+	  blksize = 0;  last_char = CH_NEWLINE;
+	  srcnext();
 	  startline = (p = chpt) - 1;
 	  hcount = 0;
 	}
@@ -513,18 +578,26 @@ OBJECT LexGetToken(void)
 	{ res = NewToken(CLOSURE, &file_pos, 0, 0, precedence(res), res);
 	}
 	else if( predefined(res) == INCLUDE || predefined(res) == SYS_INCLUDE )
-	{ OBJECT t, fname;  FILE_NUM fnum;  int len;
+	{ OBJECT t, fname;  FILE_NUM fnum;  int len;  BOOLEAN scope_suppressed;
 	  chpt = p;
 	  t = LexGetToken();
-	  if( type(t) != LBR )
+	  scope_suppressed = (type(t)==WORD && StringEqual(string(t), KW_LBR));
+
+	  if( type(t)!=LBR && !scope_suppressed )
 	  { Error(2, 9, "%s expected (after %s)",
 	      WARN, &fpos(t), KW_LBR, SymName(res));
 	    Dispose(t);
 	    res = nilobj;
 	    break;
 	  }
+	  if( scope_suppressed )
+	  { UnSuppressScope();
+	    Dispose(t);
+	    New(t, LBR);
+	  }
 	  fname = Parse(&t, nilobj, FALSE, FALSE);
 	  fname = ReplaceWithTidy(fname, FALSE);
+	  if( scope_suppressed ) SuppressScope();
 	  if( !is_word(type(fname)) )
 	  { Error(2, 10, "name of include file expected here",
 	      WARN, &fpos(fname));
@@ -540,7 +613,7 @@ OBJECT LexGetToken(void)
 	      INCLUDE_FILE,
 	      predefined(res)==INCLUDE ? INCLUDE_PATH : SYSINCLUDE_PATH);
 	  Dispose(fname);
-	  LexPush(fnum, 0, INCLUDE_FILE);
+	  LexPush(fnum, 0, INCLUDE_FILE, 1, FALSE);
 	  res = LexGetToken();
 	  vcount++; /** TEST ADDITION! **/
 	  p = chpt;
@@ -702,12 +775,12 @@ FILE *fp;  BOOLEAN end_stop;  FILE_POS *err_pos;
       case OTHER:
       
 	skipping = FALSE;
-	if( *(p-1) == '{' )
+	if( *(p-1) == '{' /*}*/ )
 	{ clear();
 	  putc(*(p-1), fp);
 	  depth++;
 	}
-	else if( *(p-1) == '}' )
+	else if( *(p-1) == /*{*/ '}' )
 	{ if( !end_stop && depth == 0 )
 	  { p--;
 	    finished = TRUE;
@@ -760,7 +833,7 @@ FILE *fp;  BOOLEAN end_stop;  FILE_POS *err_pos;
 	    incl_fp = OpenFile(fnum, FALSE, TRUE);
 	    if( incl_fp == NULL )
 	      Error(2, 20, "cannot open include file %s",
-		FATAL, PosOfFile(fnum), FileName(fnum));
+		FATAL, PosOfFile(fnum), FullFileName(fnum));
 	    while( (ch = getc(incl_fp)) != EOF )
 	      putc(ch, fp);
 	    fclose(incl_fp);
