@@ -1,6 +1,6 @@
 /*@z11.c:Style Service:EchoStyle()@*******************************************/
 /*                                                                           */
-/*  THE LOUT DOCUMENT FORMATTING SYSTEM (VERSION 3.08)                       */
+/*  THE LOUT DOCUMENT FORMATTING SYSTEM (VERSION 3.10)                       */
 /*  COPYRIGHT (C) 1991, 1996 Jeffrey H. Kingston                             */
 /*                                                                           */
 /*  Jeffrey H. Kingston (jeff@cs.usyd.edu.au)                                */
@@ -190,6 +190,8 @@ static void changebreak(STYLE *style, OBJECT x)
 	fill_style(*style) = FILL_ON, display_style(*style) = DISPLAY_CENTRE;
     else if( StringEqual(string(x), STR_BREAK_RRAGGED) )
 	fill_style(*style) = FILL_ON, display_style(*style) = DISPLAY_RIGHT;
+    else if( StringEqual(string(x), STR_BREAK_ORAGGED) )
+	fill_style(*style) = FILL_ON, display_style(*style) = DISPLAY_ORAGGED;
     else if( StringEqual(string(x), STR_BREAK_LINES) )
 	fill_style(*style) = FILL_OFF, display_style(*style) = DISPLAY_LEFT;
     else if( StringEqual(string(x), STR_BREAK_CLINES) )
