@@ -1,6 +1,6 @@
 /*@z23.c:Galley Printer:ScaleFactor()@****************************************/
 /*                                                                           */
-/*  THE LOUT DOCUMENT FORMATTING SYSTEM (VERSION 3.08)                       */
+/*  THE LOUT DOCUMENT FORMATTING SYSTEM (VERSION 3.10)                       */
 /*  COPYRIGHT (C) 1991, 1996 Jeffrey H. Kingston                             */
 /*                                                                           */
 /*  Jeffrey H. Kingston (jeff@cs.usyd.edu.au)                                */
@@ -30,6 +30,7 @@
 #include "externs"
 #define	NO_SUPPRESS	FALSE
 #define	SUPPRESS	TRUE
+#define word_equal(x, str)  (is_word(type(x)) && StringEqual(string(x), str))
 
 #define CountChild(y, link, i)						\
 for( y=pred(link, PARENT), i=1; type(y)==LINK;  y = pred(y, PARENT), i++ )
@@ -43,7 +44,7 @@ for( y=pred(link, PARENT), i=1; type(y)==LINK;  y = pred(y, PARENT), i++ )
 /*                                                                           */
 /*****************************************************************************/
 
-static float ScaleFactor(LENGTH avail_size, LENGTH inner_size)
+static float ScaleFactor(FULL_LENGTH avail_size, FULL_LENGTH inner_size)
 { float scale_factor;
   scale_factor = avail_size <= 0 ? 0 :
 		 inner_size <= 0 ? 0 : (float) avail_size / inner_size;
@@ -53,17 +54,17 @@ static float ScaleFactor(LENGTH avail_size, LENGTH inner_size)
 
 /*@::FindAdjustIncrement()@***************************************************/
 /*                                                                           */
-/*  static LENGTH FindAdjustIncrement(x, frame_size, dim)                    */
+/*  static FULL_LENGTH FindAdjustIncrement(x, frame_size, dim)               */
 /*                                                                           */
 /*  Find the amount by which to increase the width of the subobjects of      */
 /*  concatenation object x so that it is adjusted to fill size frame_size.   */
 /*                                                                           */
 /*****************************************************************************/
 
-static LENGTH FindAdjustIncrement(OBJECT x, LENGTH frame_size, int dim)
+static FULL_LENGTH FindAdjustIncrement(OBJECT x, FULL_LENGTH frame_size,int dim)
 { OBJECT y, link, prev, g;
   int adjustable_gaps;  BOOLEAN jn;
-  LENGTH inc, mk, actual_size;
+  FULL_LENGTH inc, mk, actual_size;
 
   debug2(DGP, DD, "FindAdjustIncrement(x, %s, %s)",
 	EchoLength(frame_size), dimen(dim));
@@ -125,12 +126,12 @@ static LENGTH FindAdjustIncrement(OBJECT x, LENGTH frame_size, int dim)
 /*                                                                           */
 /*****************************************************************************/
 
-void FixAndPrintObject(OBJECT x, LENGTH xmk, LENGTH xb, LENGTH xf, int dim,
-BOOLEAN suppress, LENGTH pg, int count)
+void FixAndPrintObject(OBJECT x, FULL_LENGTH xmk, FULL_LENGTH xb,
+  FULL_LENGTH xf, int dim, BOOLEAN suppress, FULL_LENGTH pg, int count)
 { OBJECT y, link, prev, g, uplink, z, face;
-  LENGTH mk, ymk, frame_size, back_edge, yb, yf, inc, f;
+  FULL_LENGTH mk, ymk, frame_size, back_edge, yb, yf, inc, f;
   int i; float scale_factor;  BOOLEAN jn;
-  debug6(DGP, D, "[ FixAndPrintObject(%s, %s, %s,%s, %s, %s, pg )",
+  debug6(DGP, DD, "[ FixAndPrintObject(%s, %s, %s,%s, %s, %s, pg )",
     Image(type(x)), EchoLength(xmk), EchoLength(xb), EchoLength(xf),dimen(dim),
     (suppress == SUPPRESS ? "suppress" : "no_suppress"));
   debug2(DGP, DD, "  size(x) = %s,%s;  x =",
@@ -154,6 +155,7 @@ BOOLEAN suppress, LENGTH pg, int count)
     case NULL_CLOS:
     case PAGE_LABEL:
     case CROSS:
+    case FORCE_CROSS:
     
       back(x, dim) = xb;  fwd(x, dim) = xf;
       break;
@@ -179,7 +181,7 @@ BOOLEAN suppress, LENGTH pg, int count)
       }
       else
       {
-	debug3(DGP, D, "  FAPO %s %s (underline = %s)", Image(type(x)),
+	debug3(DGP, DD, "  FAPO %s %s (underline = %s)", Image(type(x)),
 	  string(x), bool(underline(x)));
 	if( string(x)[0] != '\0' )
 	{ PrintWord(x, word_save_mark(x), pg - xmk);
@@ -226,8 +228,8 @@ BOOLEAN suppress, LENGTH pg, int count)
 
 	/* recalculate the size of x as in MinSize */
 	f = FindShift(x, y, dim);
-	back(x, dim) = min(MAX_LEN, max(0, back(y, dim) + f));
-	fwd(x, dim)  = min(MAX_LEN, max(0, fwd(y, dim)  - f));
+	back(x, dim) = min(MAX_FULL_LENGTH, max(0, back(y, dim) + f));
+	fwd(x, dim)  = min(MAX_FULL_LENGTH, max(0, fwd(y, dim)  - f));
       }
       else
       {	FixAndPrintObject(y, xmk, xb, xf, dim, suppress, pg, count);
@@ -285,8 +287,8 @@ BOOLEAN suppress, LENGTH pg, int count)
           if( dim == COLM )
             FixAndPrintObject(y, xmk, xb, xf, dim, NO_SUPPRESS, pg, count);
           else if( (scale_factor = ScaleFactor(xb+xf, size(y, ROWM))) > 0 )
-          { SaveGraphicState();
-	    CoordTranslate(0, pg-(xmk-xb+(LENGTH) (back(y,ROWM)*scale_factor)));
+          { SaveGraphicState(y);
+	    CoordTranslate(0, pg-(xmk-xb+(FULL_LENGTH) (back(y,ROWM)*scale_factor)));
 	    CoordScale(1.0, scale_factor);
             FixAndPrintObject(y, 0, back(y,ROWM), fwd(y,ROWM), dim,
 	      NO_SUPPRESS, 0, count);
@@ -327,9 +329,9 @@ BOOLEAN suppress, LENGTH pg, int count)
           }
           else if( (scale_factor =
 	    ScaleFactor(bc(constraint(x))+fc(constraint(x)),size(y,COLM))) > 0 )
-          { SaveGraphicState();
+          { SaveGraphicState(y);
 	    CoordTranslate(save_mark(x) - bc(constraint(x))
-	       + (LENGTH) (back(y, COLM)*scale_factor), 0);
+	       + (FULL_LENGTH) (back(y, COLM)*scale_factor), 0);
 	    CoordScale(scale_factor, 1.0);
             FixAndPrintObject(y, xmk, xb, xf, dim, NO_SUPPRESS, pg, count);
 	    RestoreGraphicState();
@@ -368,7 +370,7 @@ BOOLEAN suppress, LENGTH pg, int count)
           { assert( fc(constraint(x)) > 0, "FAPO: vertical scale factor!" );
 	    yb = xb * SF / fc(constraint(x));
 	    yf = xf * SF / fc(constraint(x));
-	    SaveGraphicState();
+	    SaveGraphicState(y);
 	    CoordTranslate(save_mark(x), pg - xmk);
 	    CoordScale( (float) bc(constraint(x))/SF,
 	      (float) fc(constraint(x))/SF);
@@ -382,6 +384,20 @@ BOOLEAN suppress, LENGTH pg, int count)
       break;
 
 
+    case KERN_SHRINK:
+
+      CountChild(y, LastDown(x), count);
+      if( dim == COLM )
+      { FixAndPrintObject(y, xmk, back(y,dim), fwd(y,dim), dim,
+	  NO_SUPPRESS, pg, count);
+      }
+      else
+      {	FixAndPrintObject(y, xmk, xb, xf, dim, suppress, pg, count);
+        back(x, dim) = back(y, dim);  fwd(x, dim) = fwd(y, dim);
+      }
+      break;
+
+
     case ROTATE:
     
       CountChild(y, Down(x), count);
@@ -402,18 +418,18 @@ BOOLEAN suppress, LENGTH pg, int count)
           if( dim == COLM )
           { CONSTRAINT colc, rowc, yc;
             save_mark(x) = xmk;
-	    SetConstraint(colc, back(x,COLM), MAX_LEN, fwd(x,COLM));
-	    SetConstraint(rowc, back(x,ROWM), MAX_LEN, fwd(x,ROWM));
+	    SetConstraint(colc, back(x,COLM), MAX_FULL_LENGTH, fwd(x,COLM));
+	    SetConstraint(rowc, back(x,ROWM), MAX_FULL_LENGTH, fwd(x,ROWM));
 	    RotateConstraint(&yc, y, sparec(constraint(x)), &colc, &rowc,COLM);
 	    FixAndPrintObject(y, 0, bc(yc), fc(yc), COLM,NO_SUPPRESS,pg,count);
           }
           else
           { CONSTRAINT colc, rowc, yc;
-	    SaveGraphicState();
+	    SaveGraphicState(y);
 	    CoordTranslate(save_mark(x), pg - xmk);
 	    CoordRotate(sparec(constraint(x)));
-	    SetConstraint(colc, back(x,COLM), MAX_LEN, fwd(x,COLM));
-	    SetConstraint(rowc, back(x,ROWM), MAX_LEN, fwd(x,ROWM));
+	    SetConstraint(colc, back(x,COLM), MAX_FULL_LENGTH, fwd(x,COLM));
+	    SetConstraint(rowc, back(x,ROWM), MAX_FULL_LENGTH, fwd(x,ROWM));
 	    RotateConstraint(&yc, y, sparec(constraint(x)), &colc, &rowc, ROWM);
 	    FixAndPrintObject(y, 0, bc(yc), fc(yc), ROWM, NO_SUPPRESS,0,count);
 	    RestoreGraphicState();
@@ -462,13 +478,18 @@ BOOLEAN suppress, LENGTH pg, int count)
             else pre = tmp, post = nilobj;
 	    back(x, dim) = xb;
 	    fwd(x, dim)  = xf;
-            SaveGraphicState();
+
+	    SaveTranslateDefineSave(x, save_mark(x), pg - (xmk + fwd(x, ROWM)));
+	    /* ***
+            SaveGraphicState(x);
             CoordTranslate(save_mark(x), pg - (xmk + fwd(x, ROWM)));
 	    debug4(DGP, DD, "GRAPHIC ROWM calling %s,%s %s,%s",
 	      EchoLength(back(x, COLM)), EchoLength(fwd(x, COLM)),
 	      EchoLength(back(x, ROWM)), EchoLength(fwd(x, ROWM)));
             DefineGraphicNames(x);
-            SaveGraphicState();
+            SaveGraphicState(x);
+	    *** */
+
             PrintGraphicObject(pre);
             RestoreGraphicState();
             FixAndPrintObject(y, xb, xb, xf, dim, NO_SUPPRESS, xb + xf, count);
@@ -497,8 +518,10 @@ BOOLEAN suppress, LENGTH pg, int count)
 
           if( dim == COLM )
 	  { save_mark(x) = xmk;
-	    if( sparec(constraint(x)) )
-	    { face = finfo[font(save_style(x))].original_font;
+	    if( incgraphic_ok(x) )
+	    { debug2(DGP, D, "  %s (style %s)",
+		EchoObject(x), EchoStyle(&save_style(x)));
+	      face = finfo[font(save_style(x))].original_font;
 	      if( font_page(face) < font_curr_page )
 	      {	debug3(DFT, DD, "FAPO-IG: x = %s, font = %d, face = %s",
 		  string(x), font(save_style(x)), EchoObject(face));
@@ -506,7 +529,7 @@ BOOLEAN suppress, LENGTH pg, int count)
 	      }
 	    }
 	  }
-          else if( sparec(constraint(x)) )
+          else if( incgraphic_ok(x) )
 	    PrintGraphicInclude(x, save_mark(x), pg - xmk);
 	  break;
 
@@ -562,7 +585,8 @@ BOOLEAN suppress, LENGTH pg, int count)
 	else back(x, dim) = xb, fwd(x, dim) = xf;
       }
       else
-      { OBJECT start_group, zlink, m;  BOOLEAN dble_found;  LENGTH b, f, dlen;
+      { OBJECT start_group, zlink, m;  BOOLEAN dble_found;
+	FULL_LENGTH b, f, dlen;
 	start_group = nilobj;  dble_found = FALSE;  dlen = 0;
 	debug0(DGP, DD, "  groups beginning.");
 	FirstDefinite(x, link, y, jn);
@@ -661,7 +685,7 @@ BOOLEAN suppress, LENGTH pg, int count)
 
       if( dim == COLM )
       { BOOLEAN will_adjust, adjusting;
-	LENGTH actual_size,
+	FULL_LENGTH actual_size,
 	adjust_indent, frame_size, back_edge, adjust_inc, inc, adjust_sofar;
 	int adjustable_gaps, gaps_sofar;
 	BOOLEAN underlining; int underline_xstart; FONT_NUM underline_font;
@@ -686,6 +710,16 @@ BOOLEAN suppress, LENGTH pg, int count)
 
 	FirstDefinite(x, link, y, jn);
 	if( link == x )  break;  /* no definite children, nothing to print */
+
+	/*** nasty bug finder 
+	{ OBJECT ff = y;
+	debugcond1(DGP, DD, word_equal(ff, "@ReportLayout"),
+	  "FAPO(%s, COLM)", EchoObject(x));
+	debugcond1(DGP, DD, word_equal(ff, "@ReportLayout"),
+	  "  adjust_cat(x) = %s", bool(adjust_cat(x)));
+	}
+	***/
+
 	last_bad_gap = nilobj;
 	adjustable_gaps = 0;
 	back_edge = xmk - xb;
@@ -797,14 +831,14 @@ BOOLEAN suppress, LENGTH pg, int count)
 	  /* check for underlining */
 	  if( underline(prev) )
 	  {
-	    debug3(DGP, D, "  FAPO/ACAT1 underline() := %s for %s %s",
+	    debug3(DGP, DD, "  FAPO/ACAT1 underline() := %s for %s %s",
 	      bool(FALSE), Image(type(prev)), EchoObject(prev));
 	    underline(prev) = FALSE;
 	    if( !underlining )
 	    {
 	      /* underlining begins here */
 	      underlining = TRUE;
-	      debug2(DGP, D, "underlining begins at %s %s",
+	      debug2(DGP, DD, "underlining begins at %s %s",
 		Image(type(prev)), EchoObject(prev));
 	      underline_font = is_word(type(prev)) ? word_font(prev) :
 		  font(save_style(x));
@@ -813,7 +847,7 @@ BOOLEAN suppress, LENGTH pg, int count)
 	    if( !underline(g) )
 	    {
 	      /* underlining ends here */
-	      debug2(DGP, D, "underlining ends at %s %s",
+	      debug2(DGP, DD, "underlining ends at %s %s",
 		Image(type(prev)), EchoObject(prev));
 	      New(urec, UNDER_REC);
 	      back(urec, COLM) = underline_xstart;
@@ -853,17 +887,17 @@ BOOLEAN suppress, LENGTH pg, int count)
 	}
 
 	/* check for underlining */
-	debug3(DGP, D, "  underlining is %s in %s %s", bool(underline(prev)),
+	debug3(DGP, DD, "  underlining is %s in %s %s", bool(underline(prev)),
 	  Image(type(prev)), EchoObject(prev));
 	if( underline(prev) )
 	{
-	  debug3(DGP, D, "  FAPO/ACAT1 underline() := %s for %s %s",
+	  debug3(DGP, DD, "  FAPO/ACAT1 underline() := %s for %s %s",
 	    bool(FALSE), Image(type(prev)), EchoObject(prev));
 	  underline(prev) = FALSE;
 	  if( !underlining )
 	  {
 	    /* underlining begins here */
-	    debug2(DGP, D, "underlining begins at %s %s",
+	    debug2(DGP, DD, "underlining begins at %s %s",
 	      Image(type(prev)), EchoObject(prev));
 	    underlining = TRUE;
 	    underline_font = is_word(type(prev)) ? word_font(prev) :
@@ -872,7 +906,7 @@ BOOLEAN suppress, LENGTH pg, int count)
 	  }
 
 	  /* underlining must end here */
-	  debug2(DGP, D, "underlining ends at %s %s",
+	  debug2(DGP, DD, "underlining ends at %s %s",
 	    Image(type(prev)), EchoObject(prev));
 	  New(urec, UNDER_REC);
 	  back(urec, COLM) = underline_xstart;
@@ -934,6 +968,6 @@ BOOLEAN suppress, LENGTH pg, int count)
       break;
 
   } /* end switch */
-  debug2(DGP, D, "] FixAndPrintObject returning (size now %s,%s).",
+  debug2(DGP, DD, "] FixAndPrintObject returning (size now %s,%s).",
 	EchoLength(back(x, dim)), EchoLength(fwd(x, dim)));
 } /* end FixAndPrintObject */
