@@ -1,6 +1,6 @@
 /*@z41.c:Object Input-Output:AppendToFile, ReadFromFile@**********************/
 /*                                                                           */
-/*  THE LOUT DOCUMENT FORMATTING SYSTEM (VERSION 3.08)                       */
+/*  THE LOUT DOCUMENT FORMATTING SYSTEM (VERSION 3.10)                       */
 /*  COPYRIGHT (C) 1991, 1996 Jeffrey H. Kingston                             */
 /*                                                                           */
 /*  Jeffrey H. Kingston (jeff@cs.usyd.edu.au)                                */
@@ -35,23 +35,24 @@ static FILE	*last_write_fp  = null;
 
 /*****************************************************************************/
 /*                                                                           */
-/*  OBJECT ReadFromFile(fnum, pos)                                           */
+/*  OBJECT ReadFromFile(fnum, pos, lnum)                                     */
 /*                                                                           */
 /*  Read an object from file fnum starting at position pos.                  */
 /*  The object may include @Env operators defining its environment, or       */
 /*  not, but in any case ReadFromFile assumes that the correct scope is set. */
+/*  lnum is the line number of the spot you end up at when you seek to pos.  */
 /*                                                                           */
 /*****************************************************************************/
 
-OBJECT ReadFromFile(FILE_NUM fnum, long pos)
+OBJECT ReadFromFile(FILE_NUM fnum, long pos, int lnum)
 { OBJECT t, res;
 #if DEBUG_ON
   int ipos;
 #endif
   ifdebug(DPP, D, ProfileOn("ReadFromFile"));
   ifdebug(DIO, D, ipos = (int) pos);
-  debug2(DIO, D, "ReadFromFile(%s, %d)", FileName(fnum), ipos);
-  LexPush(fnum, (int) pos, DATABASE_FILE);
+  debug3(DIO, D, "ReadFromFile(%s, %d, %d)", FileName(fnum), ipos, lnum);
+  LexPush(fnum, (int) pos, DATABASE_FILE, lnum, FALSE);
   t = LexGetToken();
   if( type(t) != LBR )
   { debug1(DIO, D, "  following because type(t) = %s", Image(type(t)));
@@ -70,16 +71,71 @@ OBJECT ReadFromFile(FILE_NUM fnum, long pos)
 } /* end ReadFromFile */
 
 
+/*****************************************************************************/
+/*                                                                           */
+/*  static Optimize(x, env)                                                  */
+/*                                                                           */
+/*****************************************************************************/
+static void OptimizeParameterList(OBJECT x, OBJECT env);
+
+static void Optimize(OBJECT x, OBJECT env)
+{ OBJECT tmp;
+  if( Down(x) != x )
+  { OptimizeParameterList(x, env);
+  }
+  tmp = ParameterCheck(x, env);
+  if( tmp != nilobj )
+  { ReplaceNode(tmp, x);
+    DisposeObject(x);
+  }
+}
+
+
+/*****************************************************************************/
+/*                                                                           */
+/*  OptimizeParameterList(x, env)                                            */
+/*                                                                           */
+/*  Optimize the space required to print the parameters of x by evaluating   */
+/*  them in environment env if this is feasible.                             */
+/*                                                                           */
+/*****************************************************************************/
+
+static void OptimizeParameterList(OBJECT x, OBJECT env)
+{ OBJECT y, z, link, t, tlink;
+  assert( type(x) == CLOSURE, "OptimizeParameterList: type(x) != CLOSURE!" );
+  if( env == nilobj )  return;
+  for( link = Down(x);  link != x;  link = NextDown(link) )
+  { Child(y, link);
+    if( type(y) == PAR )
+    { Child(z, Down(y));
+      if( type(z) == CLOSURE )
+      {
+	Optimize(z, env);
+      }
+      else if( type(z) == ACAT )
+      { for( tlink = Down(z);  tlink != z;  tlink = NextDown(tlink) )
+        { Child(t, Down(tlink));
+          if( type(t) == CLOSURE )
+	    Optimize(t, env);
+        }
+      }
+    }
+  }
+} /* end OptimizeParameter */
+
+
 /*@::WriteClosure()@**********************************************************/
 /*                                                                           */
-/*  static WriteClosure(x)                                                   */
+/*  static WriteClosure(x, linecount, fnum, env)                             */
 /*                                                                           */
 /*  Write closure x to file last_write_fp, without enclosing braces and      */
 /*  without any environment attached.  If x happens to be a closure that     */
 /*  was previously read as a @Use clause, write only @LUse and the name.     */
+/*  Increment *linecount by the number of lines written.                     */
+/*  The file being written to is fnum; the environment is env (for optim.)   */
 /*                                                                           */
 /*****************************************************************************/
-static void WriteObject(OBJECT x, int outer_prec);
+static void WriteObject(OBJECT x, int outer_prec, int *linecount, FILE_NUM fnum);
 
 static BOOLEAN need_lvis(OBJECT sym)	/* true if @LVis needed before sym */
 { return !visible(sym) &&
@@ -87,18 +143,20 @@ static BOOLEAN need_lvis(OBJECT sym)	/* true if @LVis needed before sym */
 	 type(enclosing(sym)) == LOCAL;
 } /* end need_lvis */
 
-static void WriteClosure(OBJECT x)
+static void WriteClosure(OBJECT x, int *linecount, FILE_NUM fnum, OBJECT env)
 { OBJECT y, link, z, sym;
-  BOOLEAN npar_seen, name_printed;
+  BOOLEAN npar_written, name_printed;
 
   sym = actual(x);
-  if( use_invocation(sym) == x )
+  /* *** if( use_invocation(sym) == x ) *** */
+  if( use_invocation(sym) != nilobj )
   { StringFPuts(KW_LUSE, last_write_fp);
     StringFPuts(STR_SPACE, last_write_fp);
     StringFPuts(SymName(sym), last_write_fp);
   }
   else
-  { npar_seen = FALSE;  name_printed = FALSE;
+  { npar_written = FALSE;  name_printed = FALSE;
+    OptimizeParameterList(x, env);
     for( link = Down(x);  link != x;  link = NextDown(link) )
     { Child(y, link);
       if( type(y) == PAR )  switch( type(actual(y)) )
@@ -107,7 +165,7 @@ static void WriteClosure(OBJECT x)
       
 	  assert( Down(y) != y, "WriteObject/CLOSURE: LPAR!" );
 	  Child(z, Down(y));
-	  WriteObject(z, (int) precedence(sym));
+	  WriteObject(z, (int) precedence(sym), linecount, fnum);
 	  StringFPuts(STR_SPACE, last_write_fp);
 	  break;
 
@@ -125,17 +183,22 @@ static void WriteClosure(OBJECT x)
 	    name_printed = TRUE;
 	  }
 	  StringFPuts(STR_NEWLINE, last_write_fp);
+	  *linecount += 1;
 	  StringFPuts(STR_SPACE, last_write_fp);
 	  StringFPuts(STR_SPACE, last_write_fp);
-	  StringFPuts(STR_SPACE, last_write_fp);
-	  StringFPuts(SymName(actual(y)), last_write_fp);
-	  StringFPuts(STR_SPACE, last_write_fp);
+
+	  if( npar_code(actual(y)) != ' ' )
+	  { StringFPuts(STR_ESCAPE, last_write_fp);
+	    fprintf(last_write_fp, "%c", (char) npar_code(actual(y)));
+	  }
+	  else
+	  { StringFPuts(SymName(actual(y)), last_write_fp);
+	  }
+
 	  StringFPuts(KW_LBR, last_write_fp);
-	  StringFPuts(STR_SPACE, last_write_fp);
-	  WriteObject(z, NO_PREC);
-	  StringFPuts(STR_SPACE, last_write_fp);
+	  WriteObject(z, NO_PREC, linecount, fnum);
 	  StringFPuts(KW_RBR, last_write_fp);
-	  npar_seen = TRUE;
+	  npar_written = TRUE;
 	  break;
 
 
@@ -151,16 +214,27 @@ static void WriteClosure(OBJECT x)
 	    StringFPuts(SymName(sym), last_write_fp);
 	    name_printed = TRUE;
 	  }
-	  StringFPuts(npar_seen ? STR_NEWLINE : STR_SPACE, last_write_fp);
-	  if( has_body(sym) || filter(sym) != nilobj )
+	  if( npar_written )
+	  { StringFPuts(STR_NEWLINE, last_write_fp);
+	    *linecount += 1;
+	  }
+	  else
+	  { StringFPuts(STR_SPACE, last_write_fp);
+	  }
+	  if( filter(sym) != nilobj )
+	  {
+	    assert( type(z) == FILTERED, "WriteClosure:  filter!" );
+	    WriteObject(z, NO_PREC, linecount, fnum);
+	  }
+	  else if( has_body(sym) )
 	  {
 	    StringFPuts(KW_LBR, last_write_fp);
 	    StringFPuts(STR_SPACE, last_write_fp);
-	    WriteObject(z, NO_PREC);
+	    WriteObject(z, NO_PREC, linecount, fnum);
 	    StringFPuts(STR_SPACE, last_write_fp);
 	    StringFPuts(KW_RBR, last_write_fp);
 	  }
-	  else WriteObject(z, (int) precedence(sym));
+	  else WriteObject(z, (int) precedence(sym), linecount, fnum);
 	  break;
 
 
@@ -185,15 +259,17 @@ static void WriteClosure(OBJECT x)
 
 /*@::WriteObject()@***********************************************************/
 /*                                                                           */
-/*  static WriteObject(x, outer_prec)                                        */
+/*  static WriteObject(x, outer_prec, linecount, fnum)                       */
 /*                                                                           */
 /*  Write object x to file last_write_fp, assuming it is a subobject of an   */
 /*  object and the precedence of operators enclosing it is outer_prec.       */
+/*  Increment *linecount by the number of lines written.                     */
+/*  The file being written to is fnum.                                       */
 /*                                                                           */
 /*****************************************************************************/
 
-static void WriteObject(OBJECT x, int outer_prec)
-{ OBJECT link, y, gap_obj, sym, env;  FULL_CHAR *name;
+static void WriteObject(OBJECT x, int outer_prec, int *linecount, FILE_NUM fnum)
+{ OBJECT link, y, gap_obj, sym, env;  FULL_CHAR *name; int offset, lnum;
   int prec, i, last_prec;  BOOLEAN braces_needed;
   switch( type(x) )
   {
@@ -228,17 +304,23 @@ static void WriteObject(OBJECT x, int outer_prec)
 	  { assert( type(x) == ACAT, "WriteObject: Down(y) == y!" );
 	    for( i = 1;  i <= vspace(y);  i++ )
 	      StringFPuts(STR_NEWLINE, last_write_fp);
+	    *linecount += vspace(y);
 	    for( i = 1;  i <= hspace(y);  i++ )
 	      StringFPuts(STR_SPACE,  last_write_fp);
 	    last_prec = (vspace(y) + hspace(y) == 0) ? JUXTA_PREC : ACAT_PREC;
 	  }
 	  else
 	  { Child(gap_obj, Down(y));
-	    StringFPuts(type(x)==ACAT ? STR_SPACE : STR_NEWLINE, last_write_fp);
+	    if( type(x)==ACAT )
+	      StringFPuts(STR_SPACE, last_write_fp);
+	    else
+	    { StringFPuts(STR_NEWLINE, last_write_fp);
+	      *linecount += 1;
+	    }
 	    StringFPuts(EchoCatOp(type(x), mark(gap(y)), join(gap(y))),
 	      last_write_fp);
 	    if( !is_word(type(gap_obj)) || StringLength(string(gap_obj)) != 0 )
-		WriteObject(gap_obj, FORCE_PREC);
+		WriteObject(gap_obj, FORCE_PREC, linecount, fnum);
 	    StringFPuts(STR_SPACE, last_write_fp);
 	    last_prec = prec;
 	  }
@@ -253,9 +335,9 @@ static void WriteObject(OBJECT x, int outer_prec)
 				? JUXTA_PREC : ACAT_PREC;
 	    }
 	    else next_prec = prec;
-	    WriteObject(y, max(last_prec, next_prec));
+	    WriteObject(y, max(last_prec, next_prec), linecount, fnum);
 	  }
-	  else WriteObject(y, prec);
+	  else WriteObject(y, prec, linecount, fnum);
 	}
       }
       if( prec < outer_prec )  StringFPuts(KW_RBR, last_write_fp);
@@ -265,27 +347,71 @@ static void WriteObject(OBJECT x, int outer_prec)
     case ENV:
 
       if( Down(x) == x )
-      { /* do nothing */
-      }
-      else if( Down(x) == LastDown(x) )
-      {	Child(y, Down(x));
-	assert( type(y) == CLOSURE, "WriteObject: ENV/CLOSURE!" );
-	assert( LastDown(y) != y, "WriteObject: ENV/LastDown(y)!" );
-	Child(env, LastDown(y));
-	assert( type(env) == ENV, "WriteObject: ENV/env!" );
-	WriteObject(env, NO_PREC);
-	StringFPuts(KW_LBR, last_write_fp);
-	WriteClosure(y);
-	StringFPuts(KW_RBR, last_write_fp);
+      {
+	/* environment is empty */
+	StringFPuts(KW_ENVC, last_write_fp);
 	StringFPuts(STR_NEWLINE, last_write_fp);
+	*linecount += 1;
+      }
+      else if( EnvWriteRetrieve(x, fnum, &offset, &lnum) )
+      {
+	/* environment was previously written to this file */
+	StringFPuts(KW_ENVD, last_write_fp);
+	fprintf(last_write_fp, " \"%d %d\"", offset, lnum);
+	StringFPuts(STR_NEWLINE, last_write_fp);
+	*linecount += 1;
       }
       else
-      {	Child(env, LastDown(x));
-	assert( type(env) == ENV, "WriteObject: ENV/ENV!" );
-	WriteObject(env, NO_PREC);
-	Child(y, Down(x));
-	assert( type(y) == CLOSURE, "WriteObject: ENV/ENV+CLOSURE!" );
-	WriteObject(y, NO_PREC);
+      {
+	/* record the position of this environment */
+	EnvWriteInsert(x, fnum, (int) ftell(last_write_fp), *linecount);
+
+	/* write the environment */
+	if( Down(x) == LastDown(x) )
+        {
+	  /* envt contains just one closure (with its environment) */
+	  Child(y, Down(x));
+	  assert( type(y) == CLOSURE, "WriteObject: ENV/CLOSURE!" );
+	  StringFPuts(KW_LBR, last_write_fp);
+	  StringFPuts(STR_SPACE, last_write_fp);
+	  StringFPuts(KW_ENVA, last_write_fp);
+	  StringFPuts(STR_SPACE, last_write_fp);
+	  StringFPuts(KW_LBR, last_write_fp);
+	  StringFPuts(STR_SPACE, last_write_fp);
+	  WriteObject(y, NO_PREC, linecount, fnum);
+	  StringFPuts(STR_SPACE, last_write_fp);
+	  StringFPuts(KW_RBR, last_write_fp);
+	  StringFPuts(STR_SPACE, last_write_fp);
+	  StringFPuts(KW_RBR, last_write_fp);
+	  StringFPuts(STR_NEWLINE, last_write_fp);
+	  *linecount += 1;
+        }
+        else
+        {
+	  /* envt contains a closure (with envt) plus an environment */
+	  Child(env, LastDown(x));
+	  assert( type(env) == ENV, "WriteObject: ENV/ENV!" );
+	  StringFPuts(KW_LBR, last_write_fp);
+	  StringFPuts(STR_SPACE, last_write_fp);
+	  StringFPuts(KW_ENVB, last_write_fp);
+	  StringFPuts(STR_SPACE, last_write_fp);
+	  StringFPuts(KW_LBR, last_write_fp);
+	  StringFPuts(STR_SPACE, last_write_fp);
+	  WriteObject(env, NO_PREC, linecount, fnum);
+	  StringFPuts(STR_SPACE, last_write_fp);
+	  StringFPuts(KW_RBR, last_write_fp);
+	  StringFPuts(STR_NEWLINE, last_write_fp);
+	  *linecount += 1;
+	  Child(y, Down(x));
+	  assert( type(y) == CLOSURE, "WriteObject: ENV/ENV+CLOSURE!" );
+	  StringFPuts(KW_LBR, last_write_fp);
+	  WriteObject(y, NO_PREC, linecount, fnum);
+	  StringFPuts(KW_RBR, last_write_fp);
+	  StringFPuts(STR_SPACE, last_write_fp);
+	  StringFPuts(KW_RBR, last_write_fp);
+	  StringFPuts(STR_NEWLINE, last_write_fp);
+	  *linecount += 1;
+        }
       }
       break;
 
@@ -304,30 +430,39 @@ static void WriteObject(OBJECT x, int outer_prec)
 
       /* print environment */
       if( env != nilobj )
-      {	StringFPuts(KW_ENV, last_write_fp);
+      {	StringFPuts(KW_CENV, last_write_fp);
+	StringFPuts(STR_SPACE, last_write_fp);
+	StringFPuts(KW_LBR, last_write_fp);
       	StringFPuts(STR_NEWLINE, last_write_fp);
-	WriteObject(env, NO_PREC);
+	*linecount += 1;
+	WriteObject(env, NO_PREC, linecount, fnum);
+        StringFPuts(KW_RBR, last_write_fp);
+      	StringFPuts(STR_NEWLINE, last_write_fp);
+	*linecount += 1;
       }
 
       /* print left brace if needed */
       if( braces_needed )  StringFPuts(KW_LBR, last_write_fp);
 	
       /* print the closure proper */
-      WriteClosure(x);
+      WriteClosure(x, linecount, fnum, env);
 
       /* print closing brace if needed */
       if( braces_needed )  StringFPuts(KW_RBR, last_write_fp);
 
       /* print closing environment if needed */
+      /* ***
       if( env != nilobj )
-      { StringFPuts(STR_NEWLINE, last_write_fp);
-	StringFPuts(KW_CLOS, last_write_fp);
-      	StringFPuts(STR_NEWLINE, last_write_fp);
+      { StringFPuts(KW_RBR, last_write_fp);
+	StringFPuts(STR_NEWLINE, last_write_fp);
+	*linecount += 1;
       }
+      *** */
       break;
 
 
     case CROSS:
+    case FORCE_CROSS:
 
       Child(y, Down(x));
       assert( type(y) == CLOSURE, "WriteObject/CROSS: type(y) != CLOSURE!" );
@@ -337,9 +472,9 @@ static void WriteObject(OBJECT x, int outer_prec)
         StringFPuts(STR_SPACE, last_write_fp);
       }
       StringFPuts(SymName(actual(y)), last_write_fp);
-      StringFPuts(KW_CROSS, last_write_fp);
+      StringFPuts(type(x) == CROSS ? KW_CROSS : KW_FORCE_CROSS, last_write_fp);
       Child(y, LastDown(x));
-      WriteObject(y, FORCE_PREC);
+      WriteObject(y, FORCE_PREC, linecount, fnum);
       if( DEFAULT_PREC <= outer_prec )  StringFPuts(KW_RBR, last_write_fp);
       break;
 
@@ -357,6 +492,7 @@ static void WriteObject(OBJECT x, int outer_prec)
     case HCOVER:	name = KW_HCOVER;	goto SETC;
     case VCOVER:	name = KW_VCOVER;	goto SETC;
     case SCALE:		name = KW_SCALE;	goto SETC;
+    case KERN_SHRINK:	name = KW_KERN_SHRINK;	goto SETC;
     case HCONTRACT:	name = KW_HCONTRACT;	goto SETC;
     case VCONTRACT:	name = KW_VCONTRACT;	goto SETC;
     case HEXPAND:	name = KW_HEXPAND;	goto SETC;
@@ -382,6 +518,8 @@ static void WriteObject(OBJECT x, int outer_prec)
     case RUMP:		name = KW_RUMP;		goto SETC;
     case INSERT:	name = KW_INSERT;	goto SETC;
     case NEXT:		name = KW_NEXT;		goto SETC;
+    case PLUS:		name = KW_PLUS;		goto SETC;
+    case MINUS:		name = KW_MINUS;	goto SETC;
     case OPEN:		name = KW_OPEN;		goto SETC;
     case TAGGED:	name = KW_TAGGED;	goto SETC;
     case INCGRAPHIC:	name = KW_INCGRAPHIC;	goto SETC;
@@ -393,7 +531,7 @@ static void WriteObject(OBJECT x, int outer_prec)
       if( DEFAULT_PREC <= outer_prec )  StringFPuts(KW_LBR, last_write_fp);
       if( Down(x) != LastDown(x) )
       {	Child(y, Down(x));
-	WriteObject(y, DEFAULT_PREC);
+	WriteObject(y, DEFAULT_PREC, linecount, fnum);
 	StringFPuts(STR_SPACE, last_write_fp);
       }
 
@@ -406,10 +544,10 @@ static void WriteObject(OBJECT x, int outer_prec)
 	StringFPuts(STR_SPACE, last_write_fp);
 	if( type(x) == OPEN )
 	{ StringFPuts(KW_LBR, last_write_fp);
-	  WriteObject(y, NO_PREC);
+	  WriteObject(y, NO_PREC, linecount, fnum);
 	  StringFPuts(KW_RBR, last_write_fp);
 	}
-	else WriteObject(y, DEFAULT_PREC);
+	else WriteObject(y, DEFAULT_PREC, linecount, fnum);
       }
       if( DEFAULT_PREC <= outer_prec )  StringFPuts(KW_RBR, last_write_fp);
       break;
@@ -417,7 +555,7 @@ static void WriteObject(OBJECT x, int outer_prec)
 
     case FILTERED:
 
-      FilterWrite(x, last_write_fp);
+      FilterWrite(x, last_write_fp, linecount);
       break;
 
 
@@ -432,15 +570,15 @@ static void WriteObject(OBJECT x, int outer_prec)
 
 /*@::AppendToFile(), CloseFiles()@********************************************/
 /*                                                                           */
-/*  AppendToFile(x, fnum, pos)                                               */
+/*  AppendToFile(x, fnum, pos, lnum)                                         */
 /*                                                                           */
 /*  Append object x to file fnum, returning its fseek position in *pos.      */
-/*  Record the fact that this file has been updated.                         */
+/*  and its line number in lnum.  Record that this file has been updated.    */
 /*                                                                           */
 /*****************************************************************************/
 
-void AppendToFile(OBJECT x, FILE_NUM fnum, int *pos)
-{ FULL_CHAR buff[MAX_BUFF], *str;
+void AppendToFile(OBJECT x, FILE_NUM fnum, int *pos, int *lnum)
+{ FULL_CHAR buff[MAX_BUFF], *str;  int linecount;
   debug2(DIO, D, "AppendToFile( %s, %s )", EchoObject(x), FileName(fnum));
 
   /* open file fnum for writing */
@@ -461,11 +599,13 @@ void AppendToFile(OBJECT x, FILE_NUM fnum, int *pos)
   /* write x out and record the fact that fnum has changed */
   *pos = (int) ftell(last_write_fp);
   StringFPuts(KW_LBR, last_write_fp);
-  WriteObject(x, NO_PREC);
+  linecount = FileGetLineCount(fnum);
+  *lnum = linecount + 1;
+  WriteObject(x, NO_PREC, &linecount, fnum);
   StringFPuts(KW_RBR, last_write_fp);
   StringFPuts(STR_NEWLINE, last_write_fp);
   StringFPuts(STR_NEWLINE, last_write_fp);
-  FileSetUpdated(fnum);
+  FileSetUpdated(fnum, linecount + 2);
   debug0(DIO, D, "AppendToFile returning.");
 } /* end AppendToFile */
 
