@@ -1,6 +1,6 @@
 /*@z45.c:External Sort:SortFile()@********************************************/
 /*                                                                           */
-/*  THE LOUT DOCUMENT FORMATTING SYSTEM (VERSION 3.08)                       */
+/*  THE LOUT DOCUMENT FORMATTING SYSTEM (VERSION 3.10)                       */
 /*  COPYRIGHT (C) 1991, 1996 Jeffrey H. Kingston                             */
 /*                                                                           */
 /*  Jeffrey H. Kingston (jeff@cs.usyd.edu.au)                                */
@@ -62,7 +62,7 @@
 /* from system.h */
 
 #include <string.h>
-#include <stdlib.h>
+/* already included in externs #include <stdlib.h> */
 #include <ctype.h>
 
 /* end system.h */
@@ -89,7 +89,7 @@ static int linelength = 40;
 /* bcopy -- from lib-text-util/bcopy.c */
 /* bcopy.c -- copy memory.
    renamed jeff_bcopy as a portability fix by JK
-   Copy LENGTH bytes from SOURCE to DEST.  Does not null-terminate.
+   Copy length bytes from source to dest.  Does not null-terminate.
    In the public domain.
    By David MacKenzie <djm@gnu.ai.mit.edu>.  */
 
